void Init_Manifold(pybind11::module_& m);
void Init_Simplex(pybind11::module_& m);
void Init_Orthogonal(pybind11::module_& m);
void Init_Grassmann(pybind11::module_& m);
void Init_TransRotInvPointCloud(pybind11::module_& m);
